# HyperParameters

The `HyperParameters` class serves as a hyerparameter container. A `HyperParameters` instance contains information about both the search space and the current values of each hyperparameter. 

Hyperparameters can be defined inline with the model-building code that uses them. This saves you from having to write boilerplate code and helps to make the code more maintainable.

### Example: Building a Model using `HyperParameters`

```python
import kerastuner as kt
import tensorflow as tf

def build_model(hp):
    model = tf.keras.Sequential()
    for i in range(hp.Int('layers', 3, 10)):
        model.add(tf.keras.layers.Dense(
            units=hp.Int('units_' + str(i), 50, 100, step=10),
            activation=hp.Choice('act_' + str(i), ['relu', 'tanh'])))
    model.add(tf.keras.layers.Dense(1, activation='sigmoid'))
    model.compile('adam', 'binary_crossentropy', metrics=['accuracy'])
    return model

hp = kt.HyperParameters()
model = build_model(hp)
assert 'layers' in hp
assert 'units_0' in hp

# Restrict the search space.
hp = kt.HyperParameters()
hp.Fixed('layers', 5)
model = build_model(hp)
assert hp['layers'] == 5

# Reparametrize the search space.
hp = kt.HyperParameters()
hp.Int('layers', 20, 30)
model = build_model(hp)
assert hp['layers'] >= 20
```

{{autogenerated}}
